 --
   -- Cylindered
   -- This is the specialization for vehicles which have movable parts
   --
   -- @author  Stefan Geiger
   -- @date  18/08/09
   --
   -- Copyright (C) GIANTS Software GmbH, Confidential, All Rights Reserved.
   
   Cylindered = {};
   
   function Cylindered.prerequisitesPresent(specializations)
       --return SpecializationUtil.hasSpecialization(Steerable, specializations);
       return true;
   end;
   
   function Cylindered:load(xmlFile)
   
       self.setMovingToolDirty = SpecializationUtil.callSpecializationsFunction("setMovingToolDirty");
       self.updateCylinderedInitial = SpecializationUtil.callSpecializationsFunction("updateCylinderedInitial");
   
       local cylinderedHydraulicSound = getXMLString(xmlFile, "vehicle.cylinderedHydraulicSound#file");
       if cylinderedHydraulicSound ~= nil and cylinderedHydraulicSound ~= "" then
           cylinderedHydraulicSound = Utils.getFilename(cylinderedHydraulicSound, self.baseDirectory);
           self.cylinderedHydraulicSound = createSample("cylinderedHydraulicSound");
           self.cylinderedHydraulicSoundEnabled = false;
           loadSample(self.cylinderedHydraulicSound, cylinderedHydraulicSound, false);
           self.cylinderedHydraulicSoundPitch = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.cylinderedHydraulicSound#pitchOffset"), 1);
           self.cylinderedHydraulicSoundVolume = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.cylinderedHydraulicSound#volume"), 1);
           setSamplePitch(self.cylinderedHydraulicSound, self.cylinderedHydraulicSoundPitch);
       else
           self.cylinderedHydraulicSoundPitch = 1;
           self.cylinderedHydraulicSoundVolume = 1;
       end;
   
       self.activeDirtyMovingParts = {};
   
       local referenceNodes = {};
       self.movingParts = {};
       local i=0;
       while true do
           local baseName = string.format("vehicle.movingParts.movingPart(%d)", i);
           if not hasXMLProperty(xmlFile, baseName) then
               break;
           end;
   
           local node = Utils.indexToObject(self.components, getXMLString(xmlFile, baseName.."#index"));
           local referenceFrame = Utils.indexToObject(self.components, getXMLString(xmlFile, baseName.."#referenceFrame"));
           if node ~= nil and referenceFrame ~= nil then
               local entry = {};
               entry.referencePoint = Utils.indexToObject(self.components, getXMLString(xmlFile, baseName.."#referencePoint"));
               entry.node = node;
               entry.referenceFrame = referenceFrame;
               entry.invertZ = Utils.getNoNil(getXMLBool(xmlFile, baseName.."#invertZ"), false);
               entry.scaleZ = Utils.getNoNil(getXMLBool(xmlFile, baseName.."#scaleZ"), false);
               entry.limitedAxis = getXMLInt(xmlFile, baseName.."#limitedAxis");
               entry.playSound = Utils.getNoNil(getXMLBool(xmlFile, baseName.."#playSound"), true);
   
   
               if entry.referencePoint ~= nil then
                   local localReferencePoint = Utils.indexToObject(self.components, getXMLString(xmlFile, baseName.."#localReferencePoint"));
                   local refX, refY, refZ = worldToLocal(node, getWorldTranslation(entry.referencePoint));
                   if localReferencePoint ~= nil then
                       local x,y,z = worldToLocal(node, getWorldTranslation(localReferencePoint));
   
                       entry.referenceDistance = Utils.vector3Length(refX-x, refY-y, refZ-z);
                       entry.localReferencePoint = {x, y, z};
   
                       local side = y*(refZ-z) - z*(refY-y);
                       entry.localReferenceAngleSide = side;
                   else
                       entry.referenceDistance = 0;
                       entry.localReferencePoint = {refX, refY, refZ};
                   end;
                   entry.useLocalOffset = Utils.getNoNil(getXMLBool(xmlFile, baseName.."#useLocalOffset"), false);
   
                   entry.localReferenceDistance = Utils.vector2Length(entry.localReferencePoint[2], entry.localReferencePoint[3]);
   
                   Cylindered.loadTranslatingParts(self, xmlFile, baseName, entry);
               end
               entry.isDirty = false;
   
   
               if referenceNodes[node] == nil then
                   referenceNodes[node] = {};
               end;
               table.insert(referenceNodes[node], entry);
   
               Cylindered.loadDependentParts(self, xmlFile, baseName, entry);
   
               Cylindered.loadComponentJoints(self, xmlFile, baseName, entry);
               Cylindered.loadAttacherJoints(self, xmlFile, baseName, entry);
   
               table.insert(self.movingParts, entry);
   
               if Utils.getNoNil(getXMLBool(xmlFile, baseName.."#isActiveDirty"), false) then
                   table.insert(self.activeDirtyMovingParts, entry);
               end
           end;
           i = i+1;
       end;
   
       -- find dependencies
       for _, part in pairs(self.movingParts) do
           part.dependentParts = {};
           for _, ref in pairs(part.dependentPartNodes) do
               if referenceNodes[ref] ~= nil then
                   for _, p in pairs(referenceNodes[ref]) do
                       part.dependentParts[p] = p;
                   end;
               end;
           end;
       end;
   
   
       function hasDependentPart(w1, w2)
           if w1.dependentParts[w2] ~= nil then
               return true;
           else
               for _, v in pairs(w1.dependentParts) do
                   if hasDependentPart(v, w2) then
                       return true;
                   end;
               end;
           end;
           return false;
       end;
   
       -- sort moving parts by dependencies
       function movingPartsSort(w1,w2)
           if hasDependentPart(w1, w2) then
               return true;
           end
           if not hasDependentPart(w2, w1) then
               return w1.node < w2.node;
           end
           return false;
       end
   
       table.sort(self.movingParts, movingPartsSort);
   
       self.nodesToMovingTools = {};
       self.movingTools = {};
       local i=0;
       while true do
           local baseName = string.format("vehicle.movingTools.movingTool(%d)", i);
           if not hasXMLProperty(xmlFile, baseName) then
               break;
           end;
           local node = Utils.indexToObject(self.components, getXMLString(xmlFile, baseName.."#index"));
           if node ~= nil then
               local entry = {};
               entry.node = node;
   
               entry.externalRotSpeed = 0;
               entry.externalTransSpeed = 0;
               -- rotation
               local rotSpeed = getXMLFloat(xmlFile, baseName.."#rotSpeed");
               if rotSpeed ~= nil then
                   entry.rotSpeed = math.rad(rotSpeed)/1000;
               end;
               local rotAcceleration = getXMLFloat(xmlFile, baseName.."#rotAcceleration");
               if rotAcceleration ~= nil then
                   entry.rotAcceleration = math.rad(rotAcceleration)/(1000*1000);
               end;
               entry.lastRotSpeed = 0;
               local rotMax = getXMLFloat(xmlFile, baseName.."#rotMax");
               if rotMax ~= nil then
                   entry.rotMax = math.rad(rotMax);
               end;
               local rotMin = getXMLFloat(xmlFile, baseName.."#rotMin");
               if rotMin ~= nil then
                   entry.rotMin = math.rad(rotMin);
               end;
               entry.rotSendNumBits = Utils.getNoNil(getXMLInt(xmlFile, baseName.."#rotSendNumBits"), 8);
               local attachRotMax = getXMLFloat(xmlFile, baseName.."#attachRotMax");
               if attachRotMax ~= nil then
                   entry.attachRotMax = math.rad(attachRotMax);
               end;
               local attachRotMin = getXMLFloat(xmlFile, baseName.."#attachRotMin");
               if attachRotMin ~= nil then
                   entry.attachRotMin = math.rad(attachRotMin);
               end;
   
               -- translation
               local transSpeed = getXMLFloat(xmlFile, baseName.."#transSpeed");
               if transSpeed ~= nil then
                   entry.transSpeed = transSpeed/1000;
               end;
               local transAcceleration = getXMLFloat(xmlFile, baseName.."#transAcceleration");
               if transAcceleration ~= nil then
                   entry.transAcceleration = transAcceleration/(1000*1000);
               end;
               entry.lastTransSpeed = 0;
               entry.transMax = Utils.getNoNil(getXMLFloat(xmlFile, baseName.."#transMax"), 1);
               entry.transMin = Utils.getNoNil(getXMLFloat(xmlFile, baseName.."#transMin"), 0);
               entry.transSendNumBits = Utils.getNoNil(getXMLInt(xmlFile, baseName.."#transSendNumBits"), 8);
               entry.attachTransMax = getXMLFloat(xmlFile, baseName.."#attachTransMax");
               entry.attachTransMin = getXMLFloat(xmlFile, baseName.."#attachTransMin");
   
               entry.axis = getXMLString(xmlFile, baseName.."#axis");
               if entry.axis ~= nil then
                   entry.axisActionIndex = InputBinding[entry.axis];
   
                   if entry.axisActionIndex ~= nil then
                       self:addConflictCheckedInput(entry.axisActionIndex);
                   end
               end
               entry.invertAxis = Utils.getNoNil(getXMLBool(xmlFile, baseName.."#invertAxis"), false);
               entry.invertMouseAxis = Utils.getNoNil(getXMLBool(xmlFile, baseName.."#invertMouseAxis"), false);
               entry.speedFactor = Utils.getNoNil(getXMLFloat(xmlFile, baseName.."#speedFactor"), 1.0);
               entry.isDirty = false;
               entry.rotationAxis = Utils.getNoNil(getXMLInt(xmlFile, baseName.."#rotationAxis"), 1);
               entry.translationAxis = Utils.getNoNil(getXMLInt(xmlFile, baseName.."#translationAxis"), 3);
   
               entry.foldMinLimit = Utils.getNoNil(getXMLFloat(xmlFile, baseName .. "#foldMinLimit"), 0);
               entry.foldMaxLimit = Utils.getNoNil(getXMLFloat(xmlFile, baseName .. "#foldMaxLimit"), 1);
   
               local x,y,z = getRotation(node);
               entry.curRot = {x,y,z};
               local x,y,z = getTranslation(node);
               entry.curTrans = {x,y,z};
   
               entry.startRot = getXMLFloat(xmlFile, baseName.."#startRot");
               if entry.startRot ~= nil then
                   entry.startRot = math.rad(entry.startRot);
               end;
               entry.startTrans = getXMLFloat(xmlFile, baseName.."#startTrans");
   
               if referenceNodes[node] == nil then
                   referenceNodes[node] = {};
               end;
               table.insert(referenceNodes[node], entry);
   
               Cylindered.loadDependentMovingTools(self, xmlFile, baseName, entry);
   
               Cylindered.loadDependentParts(self, xmlFile, baseName, entry);
   
               Cylindered.loadComponentJoints(self, xmlFile, baseName, entry);
               Cylindered.loadAttacherJoints(self, xmlFile, baseName, entry);
   
               entry.isActive = true;
               table.insert(self.movingTools, entry);
               self.nodesToMovingTools[node] = entry;
           end;
           i = i+1;
       end;
   
       for _, part in pairs(self.movingTools) do
           part.dependentParts = {};
           for _, ref in pairs(part.dependentPartNodes) do
               if referenceNodes[ref] ~= nil then
                   for _, p in pairs(referenceNodes[ref]) do
                       part.dependentParts[p] = p;
                   end;
               end;
           end;
           for _, dependentTool in pairs(part.dependentMovingTools) do
               dependentTool.movingTool = self.nodesToMovingTools[dependentTool.node];
           end;
       end;
   
       for _, tool in pairs(self.movingTools) do
           if tool.startRot ~= nil then
               tool.curRot[tool.rotationAxis] = tool.startRot;
               setRotation(tool.node, unpack(tool.curRot));
           end;
           if tool.startTrans ~= nil then
               tool.curTrans[tool.translationAxis] = tool.startTrans;
               setTranslation(tool.node, unpack(tool.curTrans));
           end;
           Cylindered.setDirty(self, tool);
       end;
   
       self.cylinderedDirtyFlag = self:getNextDirtyFlag();
   
       for _, tool in pairs(self.movingTools) do
           if (tool.rotSpeed ~= nil or tool.transSpeed ~= nil) and tool.axisActionIndex ~= nil then
               self.isSelectable = true;
               break;
           end
       end
   
   end;
   
   function Cylindered:delete()
       if self.cylinderedHydraulicSound ~= nil then
           delete(self.cylinderedHydraulicSound);
           self.cylinderedHydraulicSound = nil;
       end;
   end;
   
   function Cylindered:readStream(streamId, connection)
       for i=1, table.getn(self.movingTools) do
           local tool = self.movingTools[i];
           local changed = false;
           if tool.transSpeed ~= nil then
               local newTrans=streamReadFloat32(streamId);
               if math.abs(newTrans - tool.curTrans[tool.translationAxis]) > 0.0001 then
                   tool.curTrans[tool.translationAxis] = newTrans;
                   setTranslation(tool.node, unpack(tool.curTrans));
                   changed = true;
               end;
           end;
           if tool.rotSpeed ~= nil then
               local newRot=streamReadFloat32(streamId)
               if math.abs(newRot - tool.curRot[tool.rotationAxis]) > 0.0001 then
                   tool.curRot[tool.rotationAxis] = newRot;
                   setRotation(tool.node, unpack(tool.curRot));
                   changed = true;
               end;
           end;
           if changed then
               Cylindered.setDirty(self, tool);
           end;
       end;
   end;
   
   function Cylindered:writeStream(streamId, connection)
       for i=1, table.getn(self.movingTools) do
           local tool = self.movingTools[i];
           if tool.transSpeed ~= nil then
               streamWriteFloat32(streamId, tool.curTrans[tool.translationAxis]);
           end;
           if tool.rotSpeed ~= nil then
               streamWriteFloat32(streamId, tool.curRot[tool.rotationAxis]);
           end;
       end;
   end;
   
   function Cylindered:readUpdateStream(streamId, timestamp, connection)
       local hasUpdate = streamReadBool(streamId);
       if hasUpdate then
           for i=1, table.getn(self.movingTools) do
               local tool = self.movingTools[i];
               if tool.axisActionIndex ~= nil then
                   local changed = false;
                   if tool.transSpeed ~= nil then
                       local newTrans=Utils.readCompressedRange(streamId, tool.transMin, tool.transMax, tool.transSendNumBits);
                       if math.abs(newTrans - tool.curTrans[tool.translationAxis]) > 0.0001 then
                           tool.curTrans[tool.translationAxis] = newTrans;
                           setTranslation(tool.node, unpack(tool.curTrans));
                           changed = true;
                       end;
                   end;
                   if tool.rotSpeed ~= nil then
                       local newRot;
                       if tool.rotMin == nil or tool.rotMax == nil then
                           newRot = Utils.readCompressedAngle(streamId);
                       else
                           newRot = Utils.readCompressedRange(streamId, tool.rotMin, tool.rotMax, tool.rotSendNumBits);
                       end;
                       if math.abs(newRot - tool.curRot[tool.rotationAxis]) > 0.0001 then
                           tool.curRot[tool.rotationAxis] = newRot;
                           setRotation(tool.node, unpack(tool.curRot));
                           changed = true;
                       end;
                   end;
                   if changed then
                       Cylindered.setDirty(self, tool);
                   end;
               end
           end;
           if not connection:getIsServer() then
               -- we are on the server, write the data to the clients
               self:raiseDirtyFlags(self.cylinderedDirtyFlag);
           end;
       end;
   end;
   
   function Cylindered:writeUpdateStream(streamId, connection, dirtyMask)
       if bitAND(dirtyMask, self.cylinderedDirtyFlag) ~= 0 and (connection:getIsServer() or connection ~= self.owner) then
           -- either we are on the client, or the target connection is not the owner
           streamWriteBool(streamId, true);
           for i=1, table.getn(self.movingTools) do
               local tool = self.movingTools[i];
               if tool.axisActionIndex ~= nil then
                   if tool.transSpeed ~= nil then
                       Utils.writeCompressedRange(streamId, tool.curTrans[tool.translationAxis], tool.transMin, tool.transMax, tool.transSendNumBits);
                   end;
                   if tool.rotSpeed ~= nil then
                       local rot = tool.curRot[tool.rotationAxis];
                       if tool.rotMin ==nil or tool.rotMax == nil then
                           Utils.writeCompressedAngle(streamId, rot);
                       else
                           Utils.writeCompressedRange(streamId, rot, tool.rotMin, tool.rotMax, tool.rotSendNumBits);
                       end;
                   end;
               end
           end;
       else
           streamWriteBool(streamId, false);
       end;
   end;
   
   
   function Cylindered.loadDependentMovingTools(self, xmlFile, baseName, entry)
       entry.dependentMovingTools = {};
       local j=0;
       while true do
           local refBaseName = baseName..string.format(".dependentMovingTool(%d)", j);
           if not hasXMLProperty(xmlFile, refBaseName) then
               break;
           end;
           local node = Utils.indexToObject(self.components, getXMLString(xmlFile, refBaseName.."#index"));
           local rotSpeedScale = getXMLFloat(xmlFile, refBaseName.."#rotSpeedScale");
           local transSpeedScale = getXMLFloat(xmlFile, refBaseName.."#transSpeedScale");
           if node ~= nil and (rotSpeedScale ~= nil or transSpeedScale ~= nil) then
               local dependentTool = {};
               dependentTool.node = node;
               dependentTool.rotSpeedScale = rotSpeedScale;
               dependentTool.transSpeedScale = transSpeedScale;
               table.insert(entry.dependentMovingTools, dependentTool);
           end;
   
           j = j+1;
       end;
   end;
   
   function Cylindered.loadDependentParts(self, xmlFile, baseName, entry)
       entry.dependentPartNodes = {};
       local j=0;
       while true do
           local refBaseName = baseName..string.format(".dependentPart(%d)", j);
           if not hasXMLProperty(xmlFile, refBaseName) then
               break;
           end;
           local node = Utils.indexToObject(self.components, getXMLString(xmlFile, refBaseName.."#index"));
           if node ~= nil then
               table.insert(entry.dependentPartNodes, node);
           end;
           j = j+1;
       end;
   end;
   
   function Cylindered.loadComponentJoints(self, xmlFile, baseName, entry)
   
       local indices = Utils.getVectorNFromString(getXMLString(xmlFile, baseName.. "#componentJointIndex"));
       local actors = Utils.getNoNil(Utils.getVectorNFromString(getXMLString(xmlFile, baseName.. "#anchorActor")), {});
       if indices ~= nil then
           local componentJoints = {};
           for i=1, table.getn(indices) do
               local componentJoint = self.componentJoints[indices[i]+1];
               if componentJoint ~= nil then
                   local anchorActor = Utils.getNoNil(actors[i], 0);
                   local jointEntry = {componentJoint=componentJoint, anchorActor=anchorActor };
   
                   local node = self.components[componentJoint.componentIndices[((anchorActor+1)%2)+1] ].node;
                   jointEntry.x,jointEntry.y,jointEntry.z = worldToLocal(componentJoint.jointNode, getWorldTranslation(node));
                   jointEntry.upX,jointEntry.upY,jointEntry.upZ = worldDirectionToLocal(componentJoint.jointNode, localDirectionToWorld(node, 0, 1, 0));
                   jointEntry.dirX,jointEntry.dirY,jointEntry.dirZ = worldDirectionToLocal(componentJoint.jointNode, localDirectionToWorld(node, 0, 0, 1));
   
                   table.insert(componentJoints, jointEntry );
               end;
           end;
           if table.getn(componentJoints) > 0 then
               entry.componentJoints = componentJoints;
           end;
       end;
   end;
   
   function Cylindered.loadAttacherJoints(self, xmlFile, baseName, entry)
   
       local indices = Utils.getVectorNFromString(getXMLString(xmlFile, baseName.. "#attacherJointIndices"));
       --local index = getXMLInt(xmlFile, baseName.. "#componentJointIndex");
       if indices ~= nil then
           local attacherJoints = {};
           for i=1, table.getn(indices) do
               local attacherJoint = self.attacherJoints[indices[i]+1];
               if attacherJoint ~= nil then
                   table.insert(attacherJoints, attacherJoint);
               end;
           end;
           if table.getn(attacherJoints) > 0 then
               entry.attacherJoints = attacherJoints;
           end;
       end;
       entry.inputAttacherJoint = Utils.getNoNil(getXMLBool(xmlFile, baseName.. "#inputAttacherJoint"), false);
   end;
   
   function Cylindered.loadTranslatingParts(self, xmlFile, baseName, entry)
       entry.translatingParts = {};
       if entry.referencePoint ~= nil then
           local j=0;
           while true do
               local refBaseName = baseName..string.format(".translatingPart(%d)", j);
               if not hasXMLProperty(xmlFile, refBaseName) then
                   break;
               end;
               local node = Utils.indexToObject(self.components, getXMLString(xmlFile, refBaseName.."#index"));
               if node ~= nil then
                   local transEntry = {};
                   transEntry.node = node;
                   local x,y,z = getTranslation(node);
                   transEntry.startPos = {x,y,z};
                   local x,y,z = worldToLocal(node, getWorldTranslation(entry.referencePoint));
                   transEntry.referenceDistance = z;
                   table.insert(entry.translatingParts, transEntry);
               end;
               j = j+1;
           end;
       end
   end;
   
   function Cylindered:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
       if not resetVehicles then
           for i, tool in ipairs(self.movingTools) do
               if tool.axisActionIndex ~= nil then
                   local toolKey = key..string.format(".movingTool%d",i);
                   local changed = false;
                   if tool.transSpeed ~= nil then
                       local newTrans = getXMLFloat(xmlFile, toolKey.."#translation");
                       if newTrans ~= nil then
                           if tool.transMax ~= nil then
                               newTrans = math.min(newTrans, tool.transMax);
                           end;
                           if tool.transMin ~= nil then
                               newTrans = math.max(newTrans, tool.transMin);
                           end;
                       end
                       if newTrans ~= nil and math.abs(newTrans - tool.curTrans[tool.translationAxis]) > 0.0001 then
                           tool.curTrans[tool.translationAxis] = newTrans;
                           setTranslation(tool.node, unpack(tool.curTrans));
                           changed = true;
                       end;
                   end;
                   if tool.rotSpeed ~= nil then
                       local newRot = getXMLFloat(xmlFile, toolKey.."#rotation");
                       if newRot ~= nil then
                           if tool.rotMax ~= nil then
                               newRot = math.min(newRot, tool.rotMax);
                           end;
                           if tool.rotMin ~= nil then
                               newRot = math.max(newRot, tool.rotMin);
                           end;
                       end
                       if newRot ~= nil and math.abs(newRot - tool.curRot[tool.rotationAxis]) > 0.0001 then
                           tool.curRot[tool.rotationAxis] = newRot;
                           setRotation(tool.node, unpack(tool.curRot));
                           changed = true;
                       end;
                   end;
                   if changed then
                       Cylindered.setDirty(self, tool);
                   end;
               end
           end
       end
       return BaseMission.VEHICLE_LOAD_OK;
   end;
   
   function Cylindered:getSaveAttributesAndNodes(nodeIdent)
       local attributes = "";
       local nodes = "";
       local numNodes = 0;
       for i, tool in ipairs(self.movingTools) do
           if tool.axisActionIndex ~= nil and (tool.transSpeed ~= nil or tool.rotSpeed ~= nil) then
               if numNodes > 0 then
                   nodes = nodes.."\n";
               end
               numNodes = numNodes + 1;
   
               nodes = nodes.. nodeIdent..string.format('<movingTool%d', i);
               if tool.transSpeed ~= nil then
                   nodes = nodes.. ' translation="'..tool.curTrans[tool.translationAxis]..'"';
               end
               if tool.rotSpeed ~= nil then
                   nodes = nodes.. ' rotation="'..tool.curRot[tool.rotationAxis]..'"';
               end
               nodes = nodes..'/>';
           end
       end
   
       return attributes, nodes;
   end;
   
   function Cylindered:mouseEvent(posX, posY, isDown, isUp, button)
   end;
   
   function Cylindered:keyEvent(unicode, sym, modifier, isDown)
   end;
   
   function Cylindered:update(dt)
   
       if self:getIsActive() then
           if self.isClient and self:getIsActiveForInput(false) and not self:hasInputConflictWithSelection() then
   
               local foldAnimTime = self.foldAnimTime;
               for i=1, table.getn(self.movingTools) do
                   local tool = self.movingTools[i];
                   local foldActive = (foldAnimTime == nil or (foldAnimTime <= tool.foldMaxLimit and foldAnimTime >= tool.foldMinLimit));
                   if tool.isActive and foldActive and (tool.rotSpeed ~= nil or tool.transSpeed ~= nil) and tool.axisActionIndex ~= nil then
                       local move, axisType = InputBinding.getInputAxis(tool.axisActionIndex);
   
                       if axisType == InputBinding.INPUTTYPE_MOUSE_AXIS then
                           if tool.invertMouseAxis then
                               move = -move;
                           end
                           move = move * tool.speedFactor;
                       else
                           if tool.invertAxis then
                               move = -move;
                           end
                       end
                       local rotSpeed = 0;
                       local transSpeed = 0;
                       if not InputBinding.isAxisZero(move) then
                           if tool.rotSpeed ~= nil then
                               rotSpeed = move*tool.rotSpeed;
                               if tool.rotAcceleration ~= nil and math.abs(rotSpeed - tool.lastRotSpeed) >= tool.rotAcceleration*dt then
                                   if rotSpeed > tool.lastRotSpeed then
                                       rotSpeed = tool.lastRotSpeed + tool.rotAcceleration*dt;
                                   else
                                       rotSpeed = tool.lastRotSpeed - tool.rotAcceleration*dt;
                                   end;
                               end;
                           end;
                           if tool.transSpeed ~= nil then
                               transSpeed = move*tool.transSpeed;
                               if tool.transAcceleration ~= nil and math.abs(transSpeed - tool.lastTransSpeed) >= tool.transAcceleration*dt then
                                   if transSpeed > tool.lastTransSpeed then
                                       transSpeed = tool.lastTransSpeed + tool.transAcceleration*dt;
                                   else
                                       transSpeed = tool.lastTransSpeed - tool.transAcceleration*dt;
                                   end;
                               end;
                           end;
                       else
                           -- decelerate
                           if tool.externalRotSpeed ~= 0 then
                               rotSpeed = tool.externalRotSpeed;
                           elseif tool.rotAcceleration ~= nil then
                               if tool.lastRotSpeed < 0 then
                                   rotSpeed = math.min(tool.lastRotSpeed + tool.rotAcceleration*dt, 0);
                               else
                                   rotSpeed = math.max(tool.lastRotSpeed - tool.rotAcceleration*dt, 0);
                               end;
                           end;
                           if tool.externalTransSpeed ~= 0 then
                               transSpeed = tool.externalTransSpeed;
                           elseif tool.transAcceleration ~= nil then
                               if tool.lastTransSpeed < 0 then
                                   transSpeed = math.min(tool.lastTransSpeed + tool.transAcceleration*dt, 0);
                               else
                                   transSpeed = math.max(tool.lastTransSpeed - tool.transAcceleration*dt, 0);
                               end;
                           end;
                       end;
   
                       local changed = false;
                       if rotSpeed ~= 0 then
                           local newRot = tool.curRot[tool.rotationAxis]+rotSpeed*dt;
                           if tool.rotMax ~= nil then
                               newRot = math.min(newRot, tool.rotMax);
                           end
                           if tool.rotMin ~= nil then
                               newRot = math.max(newRot, tool.rotMin);
                           end
                           local diff = newRot - tool.curRot[tool.rotationAxis];
                           tool.lastRotSpeed = diff/dt;
                           if math.abs(diff) > 0.0001 then
                               -- wrap if not limited
                               if tool.rotMin == nil and tool.rotMax == nil then
                                   if newRot > 2*math.pi then
                                       newRot = newRot - 2*math.pi;
                                   end;
                                   if newRot < 0 then
                                       newRot = newRot + 2*math.pi;
                                   end;
                               end
                               tool.curRot[tool.rotationAxis] = newRot;
                               setRotation(tool.node, unpack(tool.curRot));
                               changed = true;
                           end;
                       else
                           tool.lastRotSpeed = 0;
                       end;
                       if transSpeed ~= 0 then
                           local newTrans = tool.curTrans[tool.translationAxis]+transSpeed*dt;
                           if tool.transMax ~= nil then
                               newTrans = math.min(newTrans, tool.transMax);
                           end;
                           if tool.transMin ~= nil then
                               newTrans = math.max(newTrans, tool.transMin);
                           end;
                           local diff = newTrans - tool.curTrans[tool.translationAxis];
                           tool.lastTransSpeed = diff/dt;
                           if math.abs(diff) > 0.0001 then
                               tool.curTrans[tool.translationAxis] = newTrans;
                               setTranslation(tool.node, unpack(tool.curTrans));
                               changed = true;
                           end;
                       else
                           tool.lastTransSpeed = 0;
                       end;
   
                       for _, dependentTool in pairs(tool.dependentMovingTools) do
                           if dependentTool.rotSpeedScale ~= nil and tool.lastRotSpeed ~= 0 then
                               dependentTool.movingTool.externalRotSpeed = dependentTool.rotSpeedScale * tool.lastRotSpeed;
                           end
                           if dependentTool.transSpeedScale ~= nil and tool.lastTransSpeed ~= 0 then
                               dependentTool.movingTool.externalTransSpeed = dependentTool.transSpeedScale * tool.lastTransSpeed;
                           end
                       end
   
                       if changed then
                           Cylindered.setDirty(self, tool);
                           self:raiseDirtyFlags(self.cylinderedDirtyFlag);
                       end;
                   end;
                   tool.externalRotSpeed = 0;
                   tool.externalTransSpeed = 0;
               end;
           end;
   
   
   
           for _, part in pairs(self.activeDirtyMovingParts) do
               Cylindered.setDirty(self, part);
           end
       end
   
       for _, tool in pairs(self.movingTools) do
           if tool.isDirty then
               if self.isServer then
                   -- update component joint
                   Cylindered.updateComponentJoints(self, tool, false);
               end;
   
               tool.isDirty = false;
           end;
       end;
   
       for i, part in ipairs(self.movingParts) do
           if part.isDirty then
               Cylindered.updateMovingPart(self, part, false);
               if part.playSound and not self.cylinderedHydraulicSoundEnabled and self:getIsActiveForSound() then
                   self.cylinderedHydraulicSoundPartNumber = i;
                   playSample(self.cylinderedHydraulicSound, 0, self.cylinderedHydraulicSoundVolume, 0);
                   self.cylinderedHydraulicSoundEnabled = true;
               end;
           else
               if self.cylinderedHydraulicSoundEnabled and self.cylinderedHydraulicSoundPartNumber == i then
                   stopSample(self.cylinderedHydraulicSound);
                   self.cylinderedHydraulicSoundEnabled = false;
               end;
           end;
       end;
   end;
   
   function Cylindered:updateTick(dt)
   end;
   
   function Cylindered:draw()
   end;
   
   function Cylindered:setMovingToolDirty(node)
       local tool = self.nodesToMovingTools[node];
       if tool ~= nil then
           local x,y,z = getRotation(tool.node);
           tool.curRot[1] = x;
           tool.curRot[2] = y;
           tool.curRot[3] = z;
           local x,y,z = getTranslation(tool.node);
           tool.curTrans[1] = x;
           tool.curTrans[2] = y;
           tool.curTrans[3] = z;
           Cylindered.setDirty(self, tool);
       end;
   end;
   
   function Cylindered.setDirty(self, part)
       if not part.isDirty then
           part.isDirty = true;
   
           if self.isServer then
               if part.attacherJoints ~= nil then
                   for _,joint in ipairs(part.attacherJoints) do
                       if joint.jointIndex ~= 0 then
                           setJointFrame(joint.jointIndex, 0, joint.jointTransform);
                       end;
                       --joint.jointFrameInvalid = true;
                   end;
               end;
               if part.inputAttacherJoint and self.attacherVehicle ~= nil then
                   local implement = self.attacherVehicle:getImplementByObject(self);
                   if implement ~= nil then
                       local jointDesc = self.attacherVehicle.attacherJoints[implement.jointDescIndex];
                       if jointDesc.jointIndex ~= 0 then
                           setJointFrame(jointDesc.jointIndex, 1, self.attacherJoint.node);
                       end
                   end
               end
           end;
   
           for _, v in pairs(part.dependentParts) do
               Cylindered.setDirty(self, v);
           end;
       end;
   end;
   
   function Cylindered.updateMovingPart(self, part, isInitialUpdate)
   
       -- the local reference point must be referenceDistance away from the referencePoint
       local refX,refY,refZ;
       local dirX, dirY, dirZ = 0,0,0;
       if part.referencePoint ~= nil then
           refX,refY,refZ = getWorldTranslation(part.referencePoint);
           if part.referenceDistance == 0 then
               if part.useLocalOffset then
                   local lx, ly, lz = worldToLocal(part.node, refX, refY, refZ);
                   dirX, dirY, dirZ = localDirectionToWorld(part.node, lx-part.localReferencePoint[1], ly-part.localReferencePoint[2], lz);
               else
                   local x,y,z = getWorldTranslation(part.node);
                   dirX, dirY, dirZ = refX - x, refY-y, refZ-z;
               end
           else
               local r1 = part.localReferenceDistance;
               local r2 = part.referenceDistance;
               local lx, ly, lz = worldToLocal(part.node, refX, refY, refZ);
               --print("intersect: "..ly .. " "..lz);
               local ix, iy, i2x, i2y = Utils.getCircleCircleIntersection(0,0, r1, ly, lz, r2);
   
               if ix ~= nil then
                   if i2x ~= nil then
                       -- use the point which as the same angle side as the original configuration
                       local side = ix*(lz-iy) - iy*(ly-ix);
                       if (side < 0) ~= (part.localReferenceAngleSide < 0) then
                           iy = i2y;
                           ix = i2x;
                       end
                   end
                   dirX, dirY, dirZ = localDirectionToWorld(part.node, 0, ix, iy)
               end;
           end;
       else
           dirX, dirY, dirZ = localDirectionToWorld(part.referenceFrame, 0, 0, 1)
       end
       if dirX ~= 0 or dirY ~= 0 or dirZ ~= 0 then
           local upX, upY, upZ = localDirectionToWorld(part.referenceFrame, 0, 1, 0);
           if part.invertZ then
               dirX = -dirX;
               dirY = -dirY;
               dirZ = -dirZ;
           end;
   
           Utils.setWorldDirection(part.node, dirX, dirY, dirZ, upX, upY, upZ, part.limitedAxis);
   
           if part.scaleZ and part.localReferenceDistance ~= nil then
               local len = Utils.vector3Length(dirX, dirY, dirZ);
               setScale(part.node, 1, 1, len/part.localReferenceDistance);
           end
       end
   
       if part.referencePoint ~= nil then
           local numTranslatingParts = table.getn(part.translatingParts)
           if numTranslatingParts > 0 then
               local _, _, dist = worldToLocal(part.node, refX, refY, refZ);
               for i=1,numTranslatingParts do
                   local translatingPart = part.translatingParts[i];
                   local newZ = (dist - translatingPart.referenceDistance)/numTranslatingParts;
                   setTranslation(translatingPart.node, translatingPart.startPos[1], translatingPart.startPos[2], newZ);
               end
           end
       end
   
       -- update component joint
       if self.isServer then
           Cylindered.updateComponentJoints(self, part, isInitialUpdate);
       end
   
       part.isDirty = false;
   end;
   
   function Cylindered.updateComponentJoints(self, entry, isInitialUpdate)
       if self.isServer then
           if entry.componentJoints ~= nil then
               for _,joint in ipairs(entry.componentJoints) do
                   local componentJoint = joint.componentJoint;
   
                   if isInitialUpdate then
   
                       local node = self.components[componentJoint.componentIndices[((joint.anchorActor+1)%2)+1] ].node;
                       local x,y,z = localToWorld(componentJoint.jointNode, joint.x, joint.y, joint.z);
                       local upX,upY,upZ = localDirectionToWorld(componentJoint.jointNode, joint.upX,joint.upY,joint.upZ);
                       local dirX,dirY,dirZ = localDirectionToWorld(componentJoint.jointNode, joint.dirX,joint.dirY,joint.dirZ);
                       Utils.setWorldTranslation(node, x,y,z);
                       Utils.setWorldDirection(node, dirX,dirY,dirZ, upX,upY,upZ);
                   end
                   setJointFrame(componentJoint.jointIndex, joint.anchorActor, componentJoint.jointNode);
               end;
           end;
       end;
   end;
   
   function Cylindered:updateCylinderedInitial()
       for _, part in pairs(self.activeDirtyMovingParts) do
           Cylindered.setDirty(self, part);
       end
   
       for _, tool in pairs(self.movingTools) do
           if tool.isDirty then
               if self.isServer then
                   Cylindered.updateComponentJoints(self, tool, true);
               end;
               tool.isDirty = false;
           end;
       end;
   
       for i, part in ipairs(self.movingParts) do
           if part.isDirty then
               Cylindered.updateMovingPart(self, part, true);
               part.isDirty = false;
           end
       end
   end
   
   function Cylindered:onAttach(attacherVehicle)
       for i=1, table.getn(self.movingTools) do
           local tool = self.movingTools[i];
           local changed = false;
           if tool.transSpeed ~= nil then
               local trans = tool.curTrans[tool.translationAxis];
   
               local changedTrans = false;
               if tool.attachTransMax ~= nil and trans > tool.attachTransMax then
                   trans = tool.attachTransMax;
                   changedTrans = true;
               elseif tool.attachTransMin ~= nil and trans < tool.attachTransMin then
                   trans = tool.attachTransMin;
                   changedTrans = true;
               end;
               if changedTrans then
                   tool.curTrans[tool.translationAxis] = trans;
                   setTranslation(tool.node, unpack(tool.curTrans));
                   changed = true;
               end;
           end;
           if tool.rotSpeed ~= nil then
               local rot = tool.curRot[tool.rotationAxis];
   
               local changedRot = false;
               if tool.attachRotMax ~= nil and rot > tool.attachRotMax then
                   rot = tool.attachRotMax;
                   changedRot = true;
               elseif tool.attachRotMin ~= nil and rot < tool.attachRotMin then
                   rot = tool.attachRotMin;
                   changedRot = true;
               end;
               if changedRot then
                   tool.curRot[tool.rotationAxis] = rot;
                   setRotation(tool.node, unpack(tool.curRot));
                   changed = true;
               end;
           end;
           if changed then
               Cylindered.setDirty(self, tool);
           end;
       end;
   end;
   
   function Cylindered:onLeave()
       if self.deactivateOnLeave then
           Cylindered.onDeactivate(self);
       else
           Cylindered.onDeactivateSounds(self);
       end;
   end;
   
   function Cylindered:onDeactivate()
       Cylindered.onDeactivateSounds(self);
   end;
   
   function Cylindered:onDeactivateSounds()
       if self.cylinderedHydraulicSoundEnabled then
           stopSample(self.cylinderedHydraulicSound);
           self.cylinderedHydraulicSoundEnabled = false;
       end;
   end;